from os import path
from math import *
from time import sleep
import numpy.polynomial.polynomial as poly

from epics import caget, caput

from iexcode.instruments.userCalcs import userCalcOut_clear
from iexcode.instruments.utilities import dateandtime, print_warning_message, read_dict
from iexcode.instruments.VLS_PGM import mono_grating_get
from iexcode.instruments.shutters import main_shutter_open

IDcal_path="/home/beams22/29IDUSER/Documents/User_Macros/Macros_29id/IEX_Dictionaries"

##############################################################################################################
################################             ID limits and calibration             ##############################
##############################################################################################################

def ID_calc_SP(mono_grating,ID_mode,hv_eV,QP_ratio):    # Mode = state (0=RCP,1=LCP,2=V,3=H)
    """Calculate the ID SP for a given polarization mode and energy;
    with Mode = 0 (RCP),1 (LCP), 2 (V), 3 (H)
    
    Previously: ID_Calc
    """
    
    
    if type(ID_mode)== str:
        ID_state=ID_state_mode(ID_mode)
    try:
        K=ID_coef(mono_grating,ID_state,hv_eV,QP_ratio)
        ID=poly.polyval(hv_eV,K)
    except KeyError:
        message_string='Not a valid ID mode!'+"\nValid Modes: "+str(ID_mode_list())
        print_warning_message(message_string)
        ID=caget("ID29:EnergySeteV")
    return round(ID,1)

##############################################################################################################
################################             ID Functions             ##############################
##############################################################################################################


def ID_wait_for_permission():
    """
    Monitors the ID permissions and waits for the ID to be in User Mode and then breaks
    Checks the status every 30 seconds

    Previously: WaitForPermission
    """
    while True:
        ID_Access=caget("ID29:AccessSecurity.VAL")
        if (ID_Access!=0):
            print("Checking ID permission, please wait..."+dateandtime())
            sleep(30)
        else:
            print("ID now in user mode -"+dateandtime())
            break

##############################################################################################################
def ID_get_all(verbose=False):
    """ 
    returns dictionary with: ID_Mode, ID_QP_ratio, ID_SP, ID_RBV

    Previously: Get_energy
    """
    vals={    
        "ID_mode":ID_mode_get(verbose=False),
        "ID_QP_ratio":ID_QP_ratio_get(verbose=False),
        "ID_sp":ID_SP_get_eV(verbose=False),
        "ID_rbv":ID_rbv_get_eV(verbose=False)
    }

    if verbose:
        print(" ID SP  : "+"%.2f" % vals["ID_sp"] , "eV    ID mode : "+vals["ID_mode"])
        print(" ID RBV : "+"%.2f" % vals['ID_rbv'], "eV    QP mode : "+str(vals['ID_QP_ratio']) +" %")
    return vals

def ID_energy_range(ID_mode=None):  
    """
    Returns the ID_min_SP, ID_max_SP for a given ID mode
    
    Previously: ID_Range
    """
    if ID_mode == None:
        ID_state=caget("ID29:ActualMode")
    else:
        ID_state = ID_state_mode(ID_mode)
    #       RCP,LCP, V , H , HN
    ID_min_SP = [400,400,440,250,250]
    ID_max_SP = 3800
    #hv_min=[390,390,430,245,245]

    return ID_min_SP[ID_state],ID_max_SP

def ID_mode_list():
    """
    returns the ID_mode_List
    current mode = 
    """
    return ["RCP", "LCP", "V", "H", "HN"]


def ID_state_mode(ID_mode):
    """ 
    Returns the state number if mode is a string 
    Returns the mode string if mode is an int or float

    Previously: ID_State2Mode(which,mode)
    """
    try:
        ID_mode = ID_mode.upper()
        ID_state =ID_mode_list().index(ID_mode)
        return ID_state

    except:
        message_string='Not a valid ID mode!'+"\nValid Modes: "+str(ID_mode_list())
        print_warning_message(message_string)

def ID_state_get():
    """
    Returns the current ID state
    """
    ID_state = caget("ID29:ActualMode")
    return ID_state

def ID_mode_get(verbose=True):
    """
    Returns the current ID mode
    """
    ID_state = caget("ID29:ActualMode")
    ID_mode = ID_mode_list()[ID_state]

    if verbose:
        print('ID mode: '+ID_mode)
    return ID_mode

def ID_mode_set(ID_mode):
    """
    writes the desired mode to the correct ID pv
    """
    caput("ID29:DesiredMode.VAL",ID_mode,wait=True,timeout=18000)     # RCP


def ID_ready(verbose=True):
    """
    check both the read and busy pv every 2 seconds

    Previously: ID_Ready
    """
    while True:
        RBV=caget("ID29:Energy.VAL")
        checkready=caget("ID29:feedback.VAL")
        checkbusy=caget("ID29:BusyRecord")
        if (checkready!="Ready") or (RBV < 3.7):
            sleep(2)
        elif ((checkready=="Ready") and (RBV > 3.7)) and (checkbusy=="Busy"):
            caput("ID29:Busy.VAL",0)
        else:
            if verbose:
                print("ID Ready")      
            return True 

def ID_power_status():
    """
    gets if the ID power supplies are On or Off
    """
    ID_OnOff=caget('ID29:Main_on_off.VAL')
    if ID_OnOff == 1: 
        return 'Off'
    elif ID_OnOff == 0:
        return 'On' 

def ID_off(verbose=True):
    """
    waits for permission then turns on the main coils OFF    
    """
    ID_wait_for_permission()
    caput("ID29:Main_on_off.VAL",1,wait=True,timeout=18000)
    sleep(5)
    if verbose:
        print("ID is now off")


def ID_on(verbose=True):
    """
    waits for permission then turns on the main coils On
    """
    ID_wait_for_permission()
    if verbose:
        print("Starting ID  -  "+dateandtime())
    
    caput("ID29:EnergySet.VAL",3.8)
    caput("ID29:Main_on_off.VAL",0,wait=True,timeout=18000)

def ID_start(ID_mode='RCP',QP_ratio=None, verbose=True):
    """
    waits for ID permission and then 
    starts ID with a specific polarization
    mode = \"H\", \"V\", \"RCP\" or \"LCP\"
    QP ratio if specified
    """
    #turns on if ID is off
    if ID_power_status =='Off':
        ID_on()
    #set QP 
    if ID_ready():
        ID_QP_mode_set(QP_ratio,verbose)
    #set ID mode
    if ID_ready():
        ID_mode_set(ID_mode, verbose)
    #opens the main shutter
    if ID_ready():
        main_shutter_open()

    if verbose:
        print('ID is now on, please set your energy')

def ID_switch_mode(ID_mode):
    """
    Change ID polarization; which = 'H', 'V', 'RCP' or 'LCP'
        if ID_mode = current mode then does nothing

    WARNING: Does not set the energy

    Previously Switch_IDMode
    """
    ID_wait_for_permission()
    main_shutter_open()
    ID_state = ID_state_get()

    try:
        if ID_state_mode(ID_mode) != ID_state:
            print("Turning ID off...")
            ID_off(verbose=True)
            sleep(10)
            
            print("Switching ID mode, please wait...")
            ID_mode_set(ID_mode) 
            ID_ready()
        print("ID Mode:",ID_mode)
    except:
        print_warning_message("Not a valid ID mode")

def ID_SP_get(verbose=False):
    """
    returns the ID setpoint in keV
    """
    ID_SP = caget("ID29:EnergyScanSet.VAL")
    if verbose:
        print("ID_SP: ", ID_SP)
    return ID_SP

def ID_SP_get_eV(verbose=False):
    """
    returns the ID setpoint in eV
    """
    ID_SP = ID_SP_get(verbose=False)
    if verbose:
        print("ID_SP: ", ID_SP)
    return ID_SP

def ID_rbv_get(verbose=False):
    """
    returns the readback value for the 
    """
    ID_RBV = caget("ID29:EnergyRBV")
    if verbose:
        print("ID_RBV: ", ID_RBV)
    return ID_RBV  

def ID_rbv_get_eV(verbose=False):
    """
    returns the readback value for the 
    """
    ID_RBV = ID_rbv_get(verbose=False)
    if verbose:
        print("ID_RBV: ", ID_RBV)
    return ID_RBV 


def ID_SP_set(hv_eV,verbose=True):
    """
    "Sets the ID set point to a specific value (hv(eV)) which will not likely be optimum"

    Previously: SetID_Raw
    """
    def _ID_write_SP_proc(keV):
        caput("ID29:EnergyScanSet.VAL",keV,wait=True,timeout=18000)
        sleep(0.5)
        caput("ID29:EnergyScanSet.VAL",(keV+0.001),wait=True,timeout=18000)
        caput('ID29:StartRamp.VAL',1)

    def _ID_bw_ok(ID_SP):
        ID_bw = ID_SP*0.095
        ID_diff = abs(ID_rbv_get()-ID_SP)
        if ID_diff > ID_bw:
            return False
        else:
            return True

    #checking permissions and opening the main shutter
    ID_wait_for_permission()
    main_shutter_open()

    #checking if desired is with allowed range, if not printing the nearest allowed value
    ID_max,ID_min = ID_energy_range()
    ID_SP=min(max(hv_eV,ID_min),ID_max)*1.0
    if hv_eV < ID_min or hv_eV > ID_max:
        message_string="Set point out of BL energy range \nPlease select a different energy."
        message_string+="\nClosest allowed value is "+str(ID_SP)
        print_warning_message(message_string)
    else:
        #desired energy is within range
        ID_SP_RBV=round(caget("ID29:EnergySet.VAL"),3)*1000
        ID_RBV=caget("ID29:EnergyRBV")
        if ID_SP == ID_SP_RBV:
            #ID is already at the SP energy
            if verbose:
                print("ID SET : "+"%.1f" % ID_SP, "eV")
                print("ID RBV : "+"%.1f" % ID_RBV, "eV")
                print(caget('ID29:TableDirection',as_string=True))
             
        else:
            # ID not at energy
            ready = ID_ready(verbose=False)
            if ready:
                #set the energy
                _ID_write_SP_proc(ID_SP/1000)

                while not ID_ready(verbose=False):
                    sleep(20)

                #is ID within some bandwidth
                if _ID_bw_ok(ID_SP) == False:
              ID_QP_ratio_get  print("QP ratio =", round(ratio,3)*100,"%")
                print("QP RBV   =", ratio_RBV,"%")    

def ID_energy_set(hv_eV,QP_ratio=None):
    """
    Sets optimum ID set point for hv(eV) (max intensity)
    and opens the main shutter

    Note that QP is generally not calibrated

    Previously: SetID
    """
    ID_mode = ID_mode_list()[ID_state_get()]
    mono_grating = mono_grating_get()
    ID_SP = ID_calc_SP(mono_grating,ID_mode,hv_eV,QP_ratio)
    ID_SP_set(ID_SP)

    
def ID_QP_ratio_get(verbose=True):
    """
    gets the read back for the QP ratio
    calculate the QP ratio

    """
    Byq=caget("ID29:ByqRdbk")
    Vcoil=caget("ID29:ByRdbk.VAL")
    ratio_calc=Byq/Vcoil
    ratio_RBV=caget("ID29:QuasiRatio.RVAL")
    if verbose:
        print("QP ratio =", round(ratio_RBV,3)*100,"%")
        print("QP RBV   =", ratio_RBV,"%")

    if abs(ratio_RBV-ratio_calc)>1:
        message_string="QP RBV and QP calc do not agree \nCheck Interlocks"
        print_warning_message(message_string)
    return ratio_RBV, ratio_calc

def ID_QP_mode_set(QP_ratio=None,verbose=True):
    """
    switched to QP mode, if not currently in
    sets the QP ratio (QP ratio min is 70) and polarization (ID mode)

    WARNING: you will need to set the polarization and mode afterward

    Previously: Switch_IDQP
    """

    #checking if above the minimum allowed value
    QP_min = 70
    if QP_ratio < QP_min:
        message_string="QP ratio is too small, setting it to minimum allowed value ("+str(QP_min)+")"
        print_warning_message(message_string)
    QP_ratio=max(70,QP_ratio)

    if QP_ratio != None:
        #checking to see if already in specified QP
        ratio_RBV, ratio_calc = ID_QP_ratio_get()
        #not the same ratio
        if ratio_RBV != QP_ratio:
            ID_off()
            caput("ID29:QuasiRatioIn.C",QP_ratio)
        
            ID_on()
            sleep(15)

            
def ID_coef(grt,ID_state,hv_eV,QP_ratio):    # Mode = state (0=RCP,1=LCP,2=V,3=H); 
    
    """Return the ID coeff for a given polarization mode and energy;
    with Mode = 0 (RCP),1 (LCP), 2 (V), 3 (H).
    Current coefficient dictionary:
        /home/beams22/29IDUSER/Documents/User_Macros/Macros_29id/IEX_Dictionaries/Dict_IDCal.txt
    
    Previously: ID_Coef
    """

    if QP_ratio == None:
        QP_ratio = ID_QP_ratio_get()[0]

    def ListRange(grt,ID_state,IDdict):  # extract the list of break pts for a given mode/grt 
        tmp_list=[]
        for item in (IDdict[grt][ID_state]):
            tmp_list.append(item[0])  
        return tmp_list


    def FindRange(hv_eV,range_list):         # returns the index for the corresponding range
        B = [x - hv_eV for x in range_list]
        #print(B)
        index = [i for (i, x) in enumerate(B) if x > 0]
        #print(index)
        return(index[0])

   
    
    if QP_ratio < 100:
        if isinstance(QP_ratio,float) or isinstance(QP_ratio,int): 
            QP_ratio=str(floor(QP_ratio))
            dictfile='Dict_IDCal_QP'+QP_ratio+'.txt'
            error_msg="calibration curve might not exist for this QP value"
        else:
            print('invalid QP value')
            return
    else:
        dictfile='Dict_IDCal.txt' 
        error_msg=""
    try:
        IDcal_fpath = path.join(IDcal_path,dictfile)
        ID_function=read_dict(IDcal_fpath)
    
    except KeyError:
        print("Unable to read dictionary "+dictfile+" ; "+error_msg) 
        
    try:   
        Lrange = ListRange(grt,ID_state,ID_function)
        Erange = FindRange(hv_eV,Lrange)
        K = ID_function[grt][ID_state][Erange][1]
        return K
        
    except KeyError:
        print("WARNING: PLease select one of the following: "+str(ID_mode_list()))


def ID_scan_pvs():
    """ 
    returns the rbv and val for scanning 
    """
    val_pv="ID29:EnergyScanSeteV"
    rbv_pv="ID29:EnergySetRBV"
    return val_pv, rbv_pv

def ID_scan_fillin(mda,scan_dim,start,stop,step,**kwargs):
    """
    fills in the scanRecord for scanning the ID set point

    **kwargs => scanRecord.fillin kwargs
    """
    #Setting up the ScanRecord for ID in Table mode
    val_pv, rbv_pv = ID_scan_pvs()
    mda.fillin(scan_dim,val_pv,rbv_pv,start,stop,step,**kwargs)


def ID_scan_fillin_table(mda,scan_dim,ID_array,**kwargs):
    """
    fills in the scanRecord for scanning the ID set point

    **kwargs => scanRecord.fillin kwargs
    """
    #Setting up the ScanRecord for ID in Table mode
    val_pv, rbv_pv = ID_scan_pvs()
    mda.fillin.table(scan_dim,val_pv,rbv_pv,ID_array,**kwargs)

##############################################################################################################
##############################             ID direction table        ##############################
##############################################################################################################

def ID_Table():
    """

    Previously: ID_Table
    """
    table = caget("ID29:TableDirection")    # up = 1 , down = 0
    By = caget("ID29:ByPolaritySet")        # pos = 1, neg = 0

    mode = ID_mode_get()

    if By > 0:
        print("\nBy > 0")
        if table == 1:            # By=1, table = 1     => -1    => A=B
            print("table = up")
            ID_direction = -1
        elif table ==  0:        # By=1, table = 0     => +1    => A#B
            print("table = down")
            ID_direction = 1
    elif By <= 0:
        print("\nBy < 0")
        if table == 1:            # By=0, table = 1     => +1    => A=B
            print("table = up")
            ID_direction = 1
        elif table ==  0:        # By=0, table = 0     => -1    => A#B
            print("table = down")
            ID_direction = -1


    if mode == "H" and mode == "RCP":
        if By > 0 and table == 0:
            print_warning_message("will do a long hysteresis if decreasing energy !!!")
#    if Mode == "HN" and Mode == "LCP":
#        if By = 0 and table == 1:
#            print "WARNING: will do a long hysteresis if decreasing energy !!!"
    print("ID direction", ID_direction)
    return ID_direction

def ID_table_userCalcOut(): 
    """
       # Work in progress

    Previously:ID_Table_CalcOut
    """
    n=4
    userCalcOut_clear("b",n)
    pvstr="29idb:userCalcOut"+str(n)
    caput(pvstr+".DESC","ID_Table")
    table="ID29:TableDirection"    # up = 1 , down = 0
    By="ID29:ByPolaritySet"        # pos = 1, neg = 0
    caput(pvstr+".INPA",table+" CP NMS")
    caput(pvstr+".INPB",By+" CP NMS")
    caput(pvstr+".CALC$","A#B")
    caput(pvstr+".OOPT","On Change")
    caput(pvstr+".DOPT","Use CALC")