from os.path import join,isfile
import time

from epics import caget, caput

from iexcode.instruments.utilities import today
import iexcode.instruments.cfg as iex

## To do: make this work if User Folder doesn't already exist
file_path = '/home/beams22/29IDUSER/Documents/User_Folders'

##############################################################################################################
##############################             logging           ##############################
##############################################################################################################
def log_print(comment=''):
    """
    prints a comment to the logfile
    """
    try:
        iex.BL.mda.log.print(comment='')
    except:
        print('No logfile written')

def log_update(d=None):
    """
    updates the log file with the last scan info
    d = dictionary of {header:[data,format]}
    """
    #try: 
    iex.BL.mda.log.update(d)

    #except:
    #    print('No logfile written')

def log_name_set(file_name):
    try: 
        iex.BL.mda.log.name_set(filename=file_name)
    except:
        print('No logfile name change')

##############################################################################################################
##############################            general   logfile functions          ##############################
##############################################################################################################
class Logfile:
    def __init__(self,endstation_name,user_name,log_dictionary,set_file=False):
        """
        endstation: used for filename pv ('ARPES' / 'Kappa' / 'Octupole' / 'Test') 
        user_name: used for filepath
        headerlist: header info

        """
        self.endstation_name = endstation_name
        self.pv = self._name_pv()
        self.userName = user_name
        self.filepath = self._fpath()
        self.log_dictionary = log_dictionary

        if set_file:
            self.name_set()
        else:
            print('\n log_filename = '+caget(self._name_pv(),as_string=True))
            print('\tuse log_name_set to change')

    def _name_pv(self):
        """
        Dictionary to get the PV in which the FileName for logging is stored
        
        endstation: 'ARPES' / 'Kappa' / 'Test

        Previously: logname_pv
        """
        #For new endstation modify here:
        pv_dict={
            'Test':'29idb:userStringSeq10.STR1',
            'ARPES':'29idb:userStringSeq10.STR2',
            'Kappa':'29idb:userStringSeq10.STR3',
            'Octupole':'29idb:userStringSeq10.STR4',
            }
        if self.endstation_name in pv_dict.keys():
            pv=pv_dict[self.endstation_name]
        else:
            pv='29idb:userStringSeq10.STR10'
        return pv

    def name_set(self,**kwargs):
        """
        Sets the string used for the FileName in scanlog and EAlog
        uses logname_PV to reference PV associated with a particular ioc
        **kwargs:
            filename = to set a specific filename, default => YYYYMMDD_log.txt

        Previously: logname_set
        """
        kwargs.setdefault('filename_suffix','_log')
        kwargs.setdefault('filename',today()+kwargs['filename_suffix']+'.txt')

        filename = kwargs['filename']

        try:
            caput(self._name_pv(),filename)
            print("\nLog filename = \'"+filename+"\' @ "+self._name_pv())
            print('To change filename, use log_name_set(new_name)')
        except:
            print("Error: was not able to set the filename, check that 29idb ioc is running")

        
    def name_get(self):
        """
        Gets the string used for the FileName in scanlog and EAlog for the given scanIOC.

        Previously: logname
        """

        return caget(self._name_pv())
       
    def _fpath(self):
        """
        returns the full path to the logfile based on the current user in the mda scanRecord


        Previously: logname, logname_generate
        """
        user_folder = join(file_path,self.userName)
        try: 
            filename = self.name_get()
            fpath_with_subfolder = join(user_folder,filename)
        except:
            fpath_with_subfolder = "logfile.txt"
            self.name_get()
            print("Couldn't read log file path, using: "+fpath_with_subfolder)
            
        return fpath_with_subfolder


    def print(self, comment=''):
        """
        Writes comments to the logfile on a new line
        comment is a string of what you want to add to the file 
            e.g. comment = "------- New Sample--------"
        
        Previously: logprint
        """

        try:
            if not isfile(self.filepath):
                self.write_header()
            with open(self.filepath,'a') as myfile:
                myfile.write(comment)
                myfile.write('\n')

        except:
            print("Logprint failed")


    def log_entries(self,endstation_dictionary=None,comment=''):
        """
        gets the meta data to write in the log file
        """
        if endstation_dictionary is None:
            endstation_dictionary = self.log_dictionary()

        scan_log = {
        "scan":(iex.BL.mda.prefix()+str(iex.BL.mda.fileNum()),"s"),
        'drive':(caget(iex.BL.mda.ioc+"scan1.P1PV"),"s"),
        'start':(caget(iex.BL.mda.ioc+"scan1.P1SP"),".3f"),
        'stop':(caget(iex.BL.mda.ioc+"scan1.P1EP"),".3f") ,
        'step':(caget(iex.BL.mda.ioc+"scan1.P1SI"),".3f"),
        }

        time_comments={
            'time': (time.strftime("%D-%H:%M:%S"),"s"),
            'comment':(comment,"s")
            }
        d = scan_log
        d.update(endstation_dictionary)
        d.update(time_comments)

        header_list = []
        data_list = []  
        format_list = []
        for key in d.keys():
            val = d[key]
            header_list.append(key)
            data_list.append(val[0])
            format_list.append(val[1])
        return header_list,data_list,format_list

    def _intit_log(self,header_list):
        """
        writes the header to the logfile
    
        Previously: SaveFile_Header
        """  
        version = '1.4'
        
        with open(self.filepath, "w+") as f:
            f.write('@Log version: '+ version+'\n\n')
            f.write('FilePath '+self.endstation_name+': '+self.filepath+'\n')
            
        label = self.endstation_name+' Header:  '
        self.write_header(header_list,label   )     

    def write_header(self,header_list,label='Header:  '): 
        """
        writes the header to the logfile
    
        Previously: SaveFile_Header
        """  
        entry = label
        with open(self.filepath, "a+") as f:
            for item in header_list:
                entry+=item+","
            f.write(entry[:-1] +'\n\n')  
    
    def write_entry(self, head_list, data_list, format_list):
        """
        writes an entry to the logfile
        """
        
        if not isfile(self.filepath):
            self.write_header(head_list)

        with open(self.filepath, "a+") as f:
            for i in range(len(format_list)):
                pv_format = "{0:"+format_list[i]+"},"
                f.write(pv_format.format(data_list[i]))
            last_entry=len(format_list)-1
            pv_format="{0:"+format_list[last_entry]+"}\n"
            f.write(pv_format.format(data_list[last_entry]))
   
    def update(self,d=None,comment=''):
        """ 
        To be used for scanlog and scanEA functions.
        Update SaveFile_Header version number when changing the structure of the file (indexing).

        Previously: SaveFile
        """
        if d == None:
            d = iex.BL.mda.log.log_dictionary()

        header_list,data_list,format_list = self.log_entries(d)

        if not isfile(self.filepath):
            self._intit_log(header_list)
        
        self.write_entry(header_list,data_list,format_list)



