"""
functions used for the mono
"""

import time
import numpy as np


from epics import caget,caput

import iexcode.instruments.cfg as iex
from iexcode.instruments.IEX_BL_config import *
from iexcode.instruments.utilities import print_warning_message,read_dict
from iexcode.instruments.scanRecord import *

##############################################################################################################
################################             mono             ##############################
##############################################################################################################
def mono_grating_names():
    """
    List of grating names 
    
    """
    ext = ['C',      'D',       'E',    'F',      'G',       'H',       'I',       "J",       "K",        "L"     ]
    GRT_names = ["MEG_Imp","HEG","MEG","Dummy",  "not used", "not used", "not used","not used","not used","not used"]
    GRT_names=GRT_names[0:3] # only use the first 3 slots
    return GRT_names

def mono_mirror_names():
    """
    List of grating names 
    
    """
    ext = ['C',      'D',       'E',    'F',      'G',       'H',       'I',       "J",       "K",        "L"     ]
    MIR_names = ["Au", "Silicon", "Carbon","not used","not used","not used","not used","not used","not used","not used"]
    MIR_names=MIR_names[0:3] # only use the first 3 slots
    return MIR_names

def _mono_extensions():
    """
    extension used for the mono mirror and grating locations
    """
    ext = ['C',      'D',       'E',    'F',      'G',       'H',       'I',       "J",       "K",        "L"     ]
    return ext

def mono_energy_range(GRT=None):
    """
    returns hv_min,hv_max, the minimum and maxium energy achievable the the specified grating
    """
    if GRT == None:
        GRT = mono_grating_get()

    if GRT == 'HEG':
        hv_min=120
        hv_max=2200
    else:
        hv_min=120
        hv_max=3000
    return hv_min,hv_max

def mono_get_all(verbose=False):
    """
    gets hv, hv_sp,grating,mirror

    Previously: Mono_Optics
    """
    d = mono_get_all_extended(verbose=False)
    vals = {
       'hv':d['energy'],
       'hv_sp':d['energy_sp'],
       'grating':d['grt'],
       'mirror':d['mir'],
    }
    if verbose:
        print(" hv: "+"%.2f" % vals['hv'], "eV, hv_sp: "+"% .2f" % vals['hv_sp'])
        print(" grating  : "+vals['grating'],", mirror  : "+vals['mirror'])
    return vals

def _mono_grating_num():
    """returns the current grating num"""
    return caget('29idmonoGRT_TYPE_MON')

def _mono_mirror_num():
    """returns the current mirror num"""
    return caget('29idmonoMIR_TYPE_MON')

def _mono_pvs(grt_num=None,mir_num=None):
    """
    returns a dictionary with all the mono pvs
    """
    if grt_num is None:
        grt_num = _mono_grating_num()
    if mir_num is None:
        mir_num = _mono_mirror_num()

    ext=_mono_extensions()

    d={
        'energy':"29idmono:ENERGY_MON",
        'energy_sp':"29idmono:ENERGY_SP",
        'grt_density':"29idmono:GRT_DENSITY",
        'grt_offset':'29idmonoGRT:P_OFFSETS.'+ext[grt_num],
        'grt_b2':'29idmonoGRT:B2_CALC.'+ext[grt_num],
        'grt_pos':'29idmonoGRT:X_DEF_POS.'+ext[grt_num],
        'grt_LD':'29idmonoGRT:TYPE_CALC.'+ext[grt_num],
        'grt_tun0':'29idmonoGRT:TUN0_CALC.'+ext[grt_num],
        'grt_tun1':'29idmonoGRT:TUN1_CALC.'+ext[grt_num],
        'grt_tun2':'29idmonoGRT:TUN2_CALC.'+ext[grt_num],
        'grt_tun3':'29idmonoGRT:TUN3_CALC.'+ext[grt_num],
        'grt_type_sp':"29idmonoGRT_TYPE_SP",
        'grt_type_move':"29idmonoGRT:X_DCPL_CALC.PROC",
        'mir_type_sp':"29idmonoMIR_TYPE_SP",
        'mir_type_move':"29idmonoMIR:X_DCPL_CALC.PROC",
        'mono_pitch_state':'29idmono:ERDY_STS',
        'grt_P_sp':"29idmonoGRT:P_SP",
        'grt_P_status':"29idmonoGRT:P_AXIS_STS",
        'grt_P_stop':"29idmonoGRT:P.STOP",
        'grt_P_kill':"29idmonoGRT:P_KILL_CMD.PROC",
        'grt_P_enable':"29idmonoGRT:P_ENA_CMD.PROC",
        'mir_P_sp':"29idmonoMIR:P_SP",
        'mir_P_status':"29idmonoMIR:P_AXIS_STS",
        'mir_P_stop':"29idmonoMIR:P.STOP",
        'mir_P_kill':"29idmonoMIR:P_KILL_CMD.PROC",
        'mir_P_enable':"29idmonoMIR:P_ENA_CMD.PROC",
        'grt_X_sp':"29idmonoGRT:X_SP",
        'grt_X_status':"29idmonoGRT:X_AXIS_STS",
        'grt_X_stop':"29idmonoGRT:X.STOP",
        'grt_X_kill':"29idmonoGRT:X_KILL_CMD.PROC",
        'grt_X_enable':"29idmonoGRT:X_ENA_CMD.PROC",
        'mir_X_sp':"29idmonoMIR:X_SP",
        'mir_X_status':"29idmonoMIR:X_AXIS_STS",
        'mir_X_stop':"29idmonoMIR:X.STOP",
        'mir_X_kill':"29idmonoMIR:X_KILL_CMD.PROC",
        'mir_X_enable':"29idmonoMIR:X_ENA_CMD.PROC",
        'mir_offset':'29idmonoMIR:P_OFFSETS.'+ext[mir_num],
        'mir_pos':'29idmonoMIR:X_DEF_POS.'+ext[mir_num],
        'mono_stop':"29idmono:STOP_CMD.PROC",
        'cff':'29idmono:CC_MON',
        'arm':'29idmono:PARAMETER.G'


    }
    return d

def mono_get_all_extended(verbose=False):
    """
    gets the mono parameters for the current grating and mirror
    returns a dictionary with those values

    Previously: Mono_Optics
    """
    grt_num = _mono_grating_num()
    mir_num = _mono_mirror_num()
    grt_names = mono_grating_names()
    mir_names = mono_mirror_names()

    ext=_mono_extensions()

    pvs = _mono_pvs()
    vals={
        'grt_num':grt_num,
        'mir_num':mir_num,
        'grt':grt_names[grt_num-1],
        'mir':mir_names[mir_num-1],
        'energy':caget(pvs['energy']),
        'energy_sp':caget(pvs['energy_sp']),
        'grt_density':caget(pvs['grt_density']),
        'grt_offset':caget(pvs['grt_offset']),
        'grt_b2':caget(pvs['grt_b2']),
        'grt_pos':caget(pvs['grt_pos']),
        'grt_LD':caget(pvs['grt_LD']),
        'grt_tun0':caget(pvs['grt_tun0']),
        'grt_tun1':caget(pvs['grt_tun1']),
        'grt_tun2':caget(pvs['grt_tun2']),
        'grt_tun3':caget(pvs['grt_tun3']),
        'grt_P_status':caget(pvs['grt_P_status']),
        'mir_P_status':caget(pvs['mir_P_status']),
        'grt_X_status':caget(pvs['grt_X_status']),
        'mir_X_status':caget(pvs['mir_X_status']),
        'mir_offset':caget(pvs['mir_offset']),
        'mir_pos':caget(pvs['mir_pos']),
        'cff':caget('29idmono:CC_MON'),
        'arm':caget('29idmono:PARAMETER.G')
        }

    if verbose:
        for key in vals.keys(): 
            print(key," = ",vals[key])

    return vals

def mono_grating_get():
    """
    returns the grating density and the string name for the grating
    """
    return mono_get_all_extended()['grt']

def mono_mirror_get():
    """
    returns the grating density and the string name for the grating
    """
    return mono_get_all_extended()['mir']

def mono_energy_get():
    """
    returns the grating density and the string name for the grating
    """
    return mono_get_all_extended()['energy']   

def mono_grating_density_get():
    """
    returns the grating density and the string name for the grating
    """
    return mono_get_all_extended()['grt_density']  

def mono_energy_set(hv_eV,verbose=True):
    """
    sets the mono energy

    Previously: SetMono
    """
    hv_min, hv_max = mono_energy_range()
    pv = _mono_pvs()['energy_sp']
    if hv_min <= hv_eV <= hv_max: 
        caput(pv,hv_eV,wait=True,timeout=60)
    time.sleep(2.5)

    mono_status = mono_status_get()
    while True:
        if mono_status > 1:
            mono_resest_pitch()
            caput(pv,hv_eV,wait=True,timeout=60)
            time.sleep(2.5)
        else:
            break
    if verbose:
        print("Mono set to",str(round(hv_eV,3)),"eV")

def mono_scan_pvs():
    """
    returns the pvs to be used in scanning
    """
    val_pv = _mono_pvs()['energy_sp']
    rbv_pv = _mono_pvs()['energy']
    return val_pv, rbv_pv

def mono_status_get():
    """
    returns the status of the mirror (GRT * MIR)
        1 = ready
        2 = moving


    Previously Mono_Status
    """
    pvs = _mono_pvs()
    mir_P_status = int(caget(pvs['mir_P_status']))
    grt_P_status = int(caget(pvs['grt_P_status']))
    mir_X_status = int(caget(pvs['mir_P_status']))
    grt_X_status = int(caget(pvs['grt_P_status']))
    mirror_status = mir_P_status * mir_X_status
    grating_status = grt_P_status * grt_X_status
    return mirror_status*grating_status

def mono_grating_set(grating_name,verbose=True):
    """
    Translate between the different grating positions
    list grating_name: mono_grating_names()

    Warning: this does not close the shutters use xrays:mirror

    Previously: Move_GRT
    """
    _mono_optic_change('grating',grating_name, verbose)

def mono_mirror_set(mirror_name,verbose=True):
    """
    Translate between the different mirror positions
    list mirror_name: mono_mirror_names()

    Warning: this does not close the shutters 

    """    
    _mono_optic_change('mirror',mirror_name, verbose)

def _mono_optic_change(optic,name, verbose=True):
    """
    optic = 'mirror','grating'
    """

    hv_eV = mono_energy_get()

    try:
        if optic == 'grating':
            num = mono_grating_names().index(name)+1
        elif optic == 'mirror':
            num = mono_mirror_names().index(name)+1
    except:
        print_warning_message(name+' not a valid '+optic)
        return

    pvs = _mono_pvs ()
    if optic == 'grating':
        type_get = mono_grating_get()
        type_sp = pvs['grt_type_sp']
        type_move = pvs['grt_type_move']
    elif optic == 'mirror':
        type_get = mono_mirror_get()
        type_sp = pvs['mir_type_sp']
        type_move = pvs['mir_type_move']
    else:
        print('Error: optic = ',optic,' not a valid choice')
        return

    if type_get != name:
        caput(type_sp,num,wait=True,timeout=18000)        
        caput(type_move,1,wait=True,timeout=18000)
        
        while True:
            if mono_status_get() > 1:
                time.sleep(5)
            else:
                break
        mono_energy_set(hv_eV)

    else:
        if verbose:
            print(optic+" is already "+name)

    if verbose:
        print("Mono "+optic+":",name)

def mono_resest_pitch():
    """
    resets the MIR and GRT pitch interlocks after a following error or limit
    Previoulsy part of SetMono
    """
    pv=_mono_pvs()
    caput(pv['grt_P_stop'],1)
    time.sleep(1)
    caput(pv['mir_P_stop'],1)
    time.sleep(1)
    print('Mono pitch was reset')

def mono_kill():
    """
    kills the mono pitch motors

    Previously: Kill_Mono
    """
    pv=_mono_pvs()
    caput(pv['grt_P_kill'],1)
    caput(pv['mir_P_kill'],1)
    caput(pv['grt_X_kill'],1)
    caput(pv['mir_X_kill'],1)    
    time.sleep(5)
    mono_stop()



def mono_stop():
    """
    presses the stop button on the mono motors 

    Previously: Stop_Mono 
    """
    pv=_mono_pvs()
    caput(pv['mono_stop'],1)
    time.sleep(5)

def mono_enable():
    """
    re-enable the mono motors 

    Previously: Enable_Mono
    """
    pv=_mono_pvs()
    caput(pv['grt_P_enable'],1)
    caput(pv['mir_P_enable'],1)
    caput(pv['grt_X_enable'],1)
    caput(pv['mir_X_enable'],1)

def mono_limits_reset():
    """
    resest the mono low limit to 200 
    was changed in the pmac to the wrong value so we need to do this

    Previously: Reset_Mono_Limits
    """

    pv=_mono_pvs()['energy_sp']
    caput(pv+".DRVL",200)
    caput(pv+".LOW",200)
    caput(pv+".LOLO",200)
    caput(pv+".LOPR",200)
    print("Mono limits have been reset.")


def mono_parameters_pv(grt_num=None,mir_num=None):
    """
    returns dictionary with mono_parameter for current grating/mirror
    """
    pvs=_mono_pvs(grt_num,mir_num)
    

    d={
        'mir_offset':pvs['mir_offset'],
        'mir_pos':pvs['mir_pos'],
        'grt_offset':pvs['grt_offset'],
        'grt_pos':pvs['grt_pos'],
        'grt_density':pvs['grt_density'],
        'grt_b2':pvs['grt_b2']
    }
    return d


def mono_parameters_get():
    """
    Gets the mono parameters and prints them for the History

    Previously: Mono_Parameters_Get
    """
    date=time.strftime("%Y%m%d",time.localtime())
    mirList = mono_mirror_names()
    grtList = mono_grating_names()
    pvList = _mono_extensions()

    #MIR
    txt="MonoParm[\'"+date+"\']= {\n\t"
    for i in range(0,len(mirList)):
        d = mono_parameters_pv(mir_num=i)
        mir = mirList[i]
        offset = caget(d['mir_offset'])
        position = caget(d['mir_pos'])
        txt+="\'"+mir+"\':["+str(offset)+","+str(position)+"],"

    #GRT
    txt+="\n\t"
    for i in range(0,len(grtList)):
        d = mono_parameters_pv(grt_num=i)
        grt = grtList[i]
        offset = caget(d['grt_offset'])
        position = caget(d['grt_pos'])
        density = caget(d['grt_density'])
        b2 = caget(d['grt_b2'])
        txt+="\'"+grt+"\':["+str(offset)+","+str(position)+","+str(density)+","+str(b2)+"],"

    txt+="}"
    print(txt)

def mono_parameters_history_read(date):
    """
    Dictionary of Mono parameters used in Mono_Parameters(date) which sets the parameters
    #MonoParm['date']={'Au':[],Si:[],'Carbon':[],'MEG_Imp':[],'HEG_JY':[],'MEG_JY':[]}
    Mono_Parameters_Set(date) writes the parameters from the Dictionary
    
    Previously: Mono_Parameters_History
    """
    #MonoParm['date']={'Au':[],Si:[],'Carbon':[],'MEG_Imp':[],'HEG_JY':[],'MEG_JY':[]}

    MonoParm=read_dict('mono_parameters.txt')
    return MonoParm[date]


def mono_parameters_history_write(date):
    """

    Previously: Mono_Parameters_Set
    """
    hv_eV=caget("29idmono:ENERGY_SP")
    MonoParm=mono_parameters_history_read(date)
    pvList=['C','D','E']
    mirList=["Au","Si","Carbon"]
    for i in range(0,len(mirList)):
        mir=mirList[i]
        caput("29idmonoMIR:P_OFFSETS."+pvList[i],MonoParm[mir][0])
        caput("29idmonoMIR:X_DEF_POS."+pvList[i],MonoParm[mir][1])
    grtList=["MEG_Imp", "HEG_JY", "MEG_JY"]
    for i in range(0,len(grtList)):
        grt=grtList[i]
        caput("29idmonoGRT:P_OFFSETS."+pvList[i],MonoParm[grt][0])
        caput("29idmonoGRT:X_DEF_POS."+pvList[i],MonoParm[grt][1])
        caput("29idmonoGRT:TYPE_CALC."+pvList[i],MonoParm[grt][2])
        caput("29idmonoGRT:B2_CALC."+pvList[i],MonoParm[grt][3])
    time.sleep (1)
    mono_energy_set(hv_eV,verbose=True)



def mono_temperature_interlock():
    """
    userCalcOut to monitor the temperatures on the Mono if the temperature gets too high then it closes the main shutter
    #29idb:userCalcOut10.VAL =0 => OK; =1 => Too hot

    Previously: Mono_TemperatureInterlock
    """
    pv='29idb:userCalcOut10'
    caput(pv+'.DESC', 'Mono Temperature Interlock')
    caput(pv+'.INPA','29idmono:TC1_MON CP NMS')
    caput(pv+'.INPB','29idmono:TC2_MON CP NMS')
    caput(pv+'.INPC','29idmono:TC3_MON CP NMS')
    caput(pv+'.INPD','29idmono:TC4_MON CP NMS')
    caput(pv+'.INPE','29idmono:TC5_MON CP NMS')
    caput(pv+'.INPF','29idmono:TC6_MON CP NMS')
    
    caput(pv+'.H','31')
    
    caput(pv+'.CALC','A>H | B>H | C>H |D>H')
    caput(pv+'OOPT','Transition To Non-zero')
    caput(pv+'.OUT','PC:29ID:FES_CLOSE_REQUEST.VAL PP NMS')

    

def mono_scan(hv_start,hv_stop,hv_step,**kwargs):
    """
    fills in the scanRecord for scanning the mono

        puts the positioner in stay after scan

        **kwargs => scanRecord.fillin kwargs
        positioner_settling_time: increased because busy record is too fast => 0.2 (default)
    """
    kwargs.setdefault('positioner_settling_time',0.2)


    #Setting up the ScanRecord for Mono in Table mode
    val_pv, rbv_pv = mono_scan_pvs()
    iex.BL.mda.fillin(val_pv,rbv_pv,hv_start,hv_stop,hv_step,**kwargs)

    #mono needs to stay and have a longer settling time
    iex.BL.mda.positioner_after_scan("STAY")

    
def mono_scan_table(hv_array,**kwargs):
    """
    fills in the scanRecord for scanning the mono

    puts the positioner in stay after scan

    **kwargs => scanRecord.fillin kwargs
    positioner_settling_time: increased because busy record is too fast => 0.2 (default)

    """
    kwargs.setdefault('positioner_settling_time',0.2)

    #Setting up the ScanRecord for Mono in Table mode
    val_pv, rbv_pv = mono_scan_pvs()
    scan_fillin_table(val_pv,rbv_pv,hv_array,**kwargs)

    #mono needs to stay and have a longer settling time
    iex.BL.mda.positioner_settling_time(**kwargs)
    iex.BL.mda.positioner_after_scan("STAY")

def mono_scan_after(**kwargs):
    """
    resets mda after scanning the mono
    """
    after_scan_pv = iex.BL.mda.default_after_scan_seq()
    caput(after_scan_pv+".PROC",1)
    iex.BL.mda.positioner_after_scan("PRIOR POS",**kwargs)


##############################################################################################################
################################             aligmnent and commissioning             ##############################
##############################################################################################################
def mono_motor_move(motor,value,wait=True,timeout=18000):
    """
    for pitch motor => MIR:P or GRT:P
    for translation motor => MIR:X or GRT:X

    Previously: Move_FMBMono
    """
    val_pv,rbv_pv = mono_motor_scan_pvs(motor)
    caput(val_pv+".PREC",3)
    caput(val_pv,value,wait=wait,timeout=timeout)

def mono_motor_scan_pvs(motor):
    """
    returns the rbv and val for scanning 
    """
    val_pv = "29idmono"+motor+"_SP"
    rbv_pv = "29idmono"+motor+"_MON"
    return val_pv,rbv_pv

def mono_motor_scan(motor,start,stop,step,**kwargs):
    """
    for pitch motor => MIR:P or GRT:P
    for translation motor => MIR:X or GRT:X

    Previously: Scan_FMBMono
    """
    val_pv,rbv_pv = mono_motor_scan_pvs(motor)
    caput(val_pv+".PREC",3) # database sets .PREC==0.  We want more digits than that.

    iex.BL.mda.fillin(val_pv,rbv_pv,start,stop,step,**kwargs)

def mono_zero_order(angle):
    """
    puts the beam in zero order => grating/mirror are parallel
    range ~1 - 5 degrees

    Previously: Mono_zero
    """
    mono_angles_set(angle,angle)
    print("Mono set to zero order:")

def mono_angles_set(alpha,beta): #JM modified to monitor the ready, moving sequentialy ended up in crash sometimes
    """
    Sets the mirror pitch (alpha) and grating pitch (beta) angles
    
    Previously: Mono_angle
    """
    alpha=alpha*1.0
    beta=beta*1.0
    pvs = _mono_pvs()

    #don't wait during move so they move in parallel (less likely to crash)
    caput(pvs['mir_P_sp'],alpha)
    caput(pvs['grt_P_sp'],beta)
    time.sleep(0.5) #need sleep before things start moving

    while caget(pvs['mono_pitch_state']) != 1:
        time.sleep(0.5)
    print("MIR_pitch = "+str(alpha)+", GRT_pitch = "+str(beta))

def mono_pink_beam():
    """
    moves grating and mirror out of beam
    x-rays will hit the pink beamstop

    Previously: Mono_pinkbeam
    """
    mono_angles_set(0,0)
    pvs=_mono_pvs()
    mono_motor_move('MIR:X',-52,wait=False)
    mono_motor_move('GRT:X',210,wait=True)

def mono_CC(): 
    """
    ##used for commissioning

    Previously: Mono_CC
    """
    caput("29idmonoMIR_TYPE_SP",3)
    caput("29idmonoMIR:X_DCPL_CALC.PROC",0)
    caput("29idmonoGRT_TYPE_SP", 10)
    caput("29idmonoGRT:X_DCPL_CALC.PROC",0)

    caput("29idmono:ENERGY_SP", 440)

def mono_grating_offset_get(grt_num=None):
    pvs=_mono_pvs(grt_num,None)
    return caget(pvs['grt_offset'])

def mono_grating_offset_set(val,grt_num=None):
    """
    sets the grating offset for the grating 

    Previously: Mono_Set_GRT0
    """
    pvs=_mono_pvs(None,None)
    caput(pvs['grt_offset'],val)
    mono_get_all_extended(verbose=True)

def mono_mirror_offset_get(mir_num=None):
    pvs=_mono_pvs(None,mir_num)
    return caget(pvs['mir_offset'])

def mono_mirror_offset_set(val,mir_num=None):
    """
    sets the mirror offset for the grating 

    Previously: Mono_Set_MIR0
    """
    pvs=_mono_pvs(None,mir_num)
    caput(pvs['mir_offset'],val)
    mono_get_all_extended(verbose=True)

def mono_grating_mirror_offsets_set(mirror_offset):
    """
    changes the mirror and grating offsets to  maintain parallelism
    After determining parallelism (zero order a=b)
    Get the correct take-off angle delta, equivalent to scan ExitSlit_Vcenter
    
    Previously: Mono_Set_MIR0_GRT0,MIR_GRT_Offset,Mono_Set_MIR0_GRT0_all
    """  
    current_vals = mono_get_all_extended(verbose=False)
    energy_sp = current_vals['energy_sp']
    mirror_offset_old = current_vals['mir_offset']
    
    #get current grating offsets
    grating_offsets=[]
    for grt_num in range(0,3):
        grating_offsets.append(mono_parameters_pv(grt_num=grt_num,mir_num=None)['grt_offset'])

    #calc mirror_offset - grating_grating    
    mir_grt_offsets = mirror_offset_old-np.array(grating_offsets)    # list of MIR-GRT for all 3 gratings

    #set mirror_offset
    mono_mirror_offset_set(mirror_offset)

    #set grating_offsets
    for grt_num in range(0,3):
        mono_grating_offset_set(mirror_offset - mir_grt_offsets[grt_num],grt_num)

    time.sleep (1)
    mono_energy_set(energy_sp)
    mono_get_all_extended(verbose=True)

def mono_grating_b2_get(grt_num=None):
    pvs = _mono_pvs(grt_num,None)
    return caget(pvs['grt_b2'])

def mono_grating_b2_set(val,grt_num=None):
    """
    sets the grating offset for the grating 
    Previously: Mono_Set_b2
    """
    hv = mono_energy_get()
    pvs = _mono_pvs(grt_num,None)
    caput(pvs['grt_b2t'],val)
    time.sleep(1)
    
    mono_energy_set(hv)
    mono_get_all_extended(verbose=True)

def mono_cff_get():
    """
    print the cff tuning parameters and arm for the current grating

    Previously: Get_CFF
    """
    d = mono_get_all_extended()
    cff = d['cff']
    tun0 = d['grt_tun0']
    tun1 = d['grt_tun1']
    tun2 = d['grt_tun2']
    tun3 = d['grt_tun3']
    arm =  d['arm']
    print(" cff : "+"%.4f" % cff , "          exit arm: "+"%.1f" % arm,"mm")
    print(" tun0 : "+"%.4e" % tun0 , "    tun1: "+"%.4e" % tun1)
    print(" tun2 : "+"%.4e" % tun2 , "    tun3: "+"%.4e" % tun3)


def mono_cff_set(val,tune_num):
    """
    sets the tuning parameters for the cff calculation 
    tune_num = order

    Previously: Mono_Set_cff
    """
    hv = mono_energy_get()
    pvs = _mono_pvs()

    #get current info
    current_vals = mono_get_all_extended(verbose=False)
    grt_pitch = current_vals['grt_pitch']
    mir_pitch = current_vals['mir_pitch']
    print ('Pitch grating, mirror: ',grt_pitch,mir_pitch)
    
    #set the tuning coeffient
    tun='tun'+str(tune_num)
    caput(pvs[tun],val)
    time.sleep(1)

    #set then energy
    mono_energy_set(hv)

    #print differences in pitch
    mono_get_all_extended(verbose=False)
    grt_dif = grt_pitch-current_vals['grt_pitch']
    mir_dif = mir_pitch-current_vals['mir_pitch']

    print('Pitch grating, mirror: ',current_vals['grt_pitch'],current_vals['mir_pitch'])
    print('Differences          : ',grt_dif,mir_dif)

def mono_arm_get():
    pvs = _mono_pvs()
    return caget(pvs(['arm']))

def mono_arm_set(distance_mm):
    """
    sets the exit arm for the grating 

    Previously: Mono_Set_ExitArm
    """
    hv = mono_energy_get()
    pvs = _mono_pvs()

    #get current info
    current_vals = mono_get_all_extended(verbose=False)
    grt_pitch = current_vals['grt_pitch']
    mir_pitch = current_vals['mir_pitch']
    print ('Pitch grating, mirror: ',grt_pitch,mir_pitch)
    
    #set the exit arm
    
    caput(pvs(['arm']),distance_mm)
    time.sleep(1)

    #set then energy
    mono_energy_set(hv)

    #print differences in pitch
    mono_get_all_extended(verbose=False)
    grt_dif = grt_pitch-current_vals['grt_pitch']
    mir_dif = mir_pitch-current_vals['mir_pitch']

    print('Pitch grating, mirror: ',current_vals['grt_pitch'],current_vals['mir_pitch'])
    print('Differences          : ',grt_dif,mir_dif)