from numpy import nan

from epics import caput, caget

import iexcode.instruments.cfg as iex
from iexcode.instruments.current_amplifiers import Keithley,SRS
from iexcode.instruments.utilities import print_warning_message
##############################################################################################################
################################            default positions             ##############################
##############################################################################################################
#For new endstation modify here:
def diagnostics_presets():
    presets={
        'H-wire':(['OUT','Start'],'29id:m1'),
        'V-wire':(['OUT','Start'],'29id:m2'),
        'H-Diagon':(['OUT','Si-2800','ML-400','ML-460'],'29id:m3'),
        'V-Diagon':(['OUT','Si-2800','ML-400','ML-460'],'29id:m4'),
        'D1A':(['OUT','Mesh'],'29id:m5'),
        'D2B':(['OUT','Mesh','YAG'],'29id:m6'),
        'D3B':(['OUT','YAG'],'29id:m7'),
        'D4C':(['OUT','Mesh','YAG'],'29id:m17'),
        'gas-cell':(['OUT','Diode1','Diode2'],'29id:m20'),
        'D4D':(['OUT','Mesh','YAG'],'29id:m25'),
        'D5D':(['OUT','Mesh','YAG','Diode'],'29id:m6')
    }
    return presets

#For new endstation modify here:
def _diagnostics_inout_dict():
    """
    dictionary to convert quick_name to preset_names
    """
    #quick_name: (diag_name, preset_in,preset_out)
    d={
        'mesh_W':('D1A','Mesh','OUT'),
        'mesh_c':('D4C','Mesh','OUT'),
        'mesh_d':('D4D','Mesh','OUT'),
        'diode_c':('gas-cell','Diode1','OUT'),
        'diode_d':('gas-D4D','Diode','OUT'),
        }

    return d

#For new endstation modify here:
def _diagnostic_read_dict():
    read_dict={
        'diode_c':Keithley('b',15),
        'mesh_d':SRS("29idMZ0:scaler1.S14", '29idd:A4'),
        'diode_d':Keithley('b',14),
    }
    return read_dict

##############################################################################################################
################################            access presets            ##############################
##############################################################################################################
def _diagnostics_preset_pv(diag_name,preset_name):
    """
    dictionary of the motor pvs for the beamline diagnostics
    """
    ioc = '29idb:'
    d = diagnostics_presets()
    pv=None
    
    if diag_name.casefold() in (key.casefold() for key in list(d.keys())):
        idx = [item.lower() for item in list(d.keys())].index(diag_name.lower())
        diag_name=list(d.keys())[idx]
        if preset_name.casefold() in (preset.casefold() for preset in d[diag_name][0]):
            idx = [item.lower() for item in d[diag_name][0]].index(preset_name.lower())
            preset_name = d[diag_name][0][idx]
            pv = ioc+diag_name+'_'+preset_name
        else:
            message ='preset_name = '+preset_name+' not a valid name'
            message +='\npreset_names:'+str(d[diag_name][0])
            print_warning_message(message)
    
    else:
        message = 'diag_name = '+diag_name+' not a valid name'
        message +='\npreset_names:'+str(list(d.keys()))
        print_warning_message(message)
    
    return pv

def diagnostics_presets_go(diag_name,preset_name,wait=False):
    pv = _diagnostics_preset_pv(diag_name,preset_name)
    caput(pv+'_Trigger.PROC',1,wait=wait,timeout=180)

def diagnostics_sp_read(diag_name,preset_name):
    pv = _diagnostics_preset_pv(diag_name,preset_name)
    caget(pv+'_Setpoint.VAL')

def _diagnostics_sp_write(diag_name,preset_name):
    pv = _diagnostics_preset_pv(diag_name,preset_name)
    caget(pv+'_Setpoint.VAL')
    

##############################################################################################################
################################            quick names            ##############################
##############################################################################################################
def diagnostic_name_list(verbose=True):
    """
    lists all diagnostics
    """
    d = _diagnostics_inout_dict()
    if verbose:
        for key in d.keys():
            print(key)
    return list(d.keys())

def diagnostic(name,in_out):
    """
    inserts/removes a diagnostic
    use 
    """
    '''
    Inserts/retracts a diagnostic either = "In" or "Out"
        => diagnostic_name_list: to list diagnostics
    '''
    d_quick = _diagnostics_inout_dict()
    if name in d_quick.keys():
        diag_name, preset_in,preset_out = d_quick[name]
        if in_out.lower() == "in":
            preset_name = preset_in
        elif in_out.lower() == "out":
            preset_name = preset_in
        diagnostics_presets_go(diag_name,preset_name)
        print("\n"+name+": "+ in_out)
    else:
        message = 'name = '+name+' is not a valid diagnostic name'
        message +='\n valid names are '+str(d_quick.keys())
        print_warning_message(message)

def diagnostics_all_out(diode_stay_in=False,mesh_stay_in=False):
    """
    Retracts all diagnostic unless
        diode_stay_in = True => leaves diode_c/diode_d In
        mesh_stay_in = True => leaves mesh_c/mesh_d In

    Previously: AllDiagOut
    """
    message = "All diagnostics out"
    d_names = _diagnostics_inout_dict()

    diag_list = list(diagnostics_presets().keys())

    #if staying remove diag_name from list
    if diode_stay_in:
        branch = iex.BL.branch
        if branch == 'c':
            diode_stay_name = d_names['diode_c'][0]
            message += '\ndiode_c => In'
        elif branch == 'd':
            diode_stay_name = d_names['diode_d'][0]
            message += '\ndiode_d => In'
        diag_list.pop(diode_stay_name) 

    if mesh_stay_in:
        branch = iex.BL.branch
        if branch == 'c':
            mesh_stay_name = d_names['mesh_c'][0]
            message += '\nmesh_c => In'
        elif branch == 'd':
            mesh_stay_name = d_names['mesh_d'][0]
            message += '\nmesh_d => In'
        diag_list.pop(mesh_stay_name) 

    #moving the diagnostics in list
    for diag_name in diag_list:
        diagnostics_presets_go(diag_name,'OUT')
    
    #putting stays in if not already in
    if diode_stay_in:
        diagnostics_presets_go(diode_stay_name,'Diode')
    if mesh_stay_name:
        diagnostics_presets_go(mesh_stay_name,'Mesh')

    print("\n",message)

def diagnostics_all_in():
    print("The follow diagnostics are in ")
    d_names = _diagnostics_inout_dict() 
    for i,name in enumerate(d_names.keys()):
        diag_name,preset_in,preset_out = d_names[name]
        if i == len(d_names.keys())-1:
            diagnostics_presets_go(diag_name,preset_in,wait=True)
        else:
            diagnostics_presets_go(diag_name,preset_in,wait=False)
        print("\t"+name)
        
def diagnostic_InOut(name,In_Out):
    d_names = _diagnostics_inout_dict() 
    try:
        diag_name,preset_in,preset_out = d_names[name] 
        if In_Out.lower() == 'in':
            preset = preset_in
        elif In_Out.lower() == 'out':
            preset = preset_out
        else:
            print('\nIn_Out = '+In_Out+' not a valid option, must be "in" or "out"')    
            return

        diagnostics_presets_go(diag_name,preset_in,wait=True)
        print("\n"+diag_name+" = "+preset)
    except:
        message = name+"not a valid diagnostic \n choose one of: "
        message += list(d_names.keys())

def mesh_W(In_Out):
    """
    Inserts/retracts the mesh in the A-hutch

    Previously MeshW
    """
    diagnostic_InOut('mesh_W',In_Out)


def diode_c(In_Out):
    """
    Inserts/retracts ARPES (gas-cell) diode

    Previously: DiodeC
    """
    diagnostic_InOut('diode_c',In_Out)


def diode_d(In_Out):
    """
    Inserts/retracts RSXS diode just before the endstation

    Previously:DiodeD
    """
    diagnostic_InOut('diode_d',In_Out)


def mesh_d(In_Out):
    """
    Inserts/retracts RSXS mesh (post-slit)
    
    Previoulsy: MeshD
    """
    diagnostic_InOut('mesh_d',In_Out)

def diagnostic_read(name, verbose=True):
    """
    reads the current amplifier and returns the value
    """ 
    try:
        det = _diagnostic_read_dict()[name]
        val=det.get(verbose)
    except:
        print_warning_message(name+" not in "+str(list(_diagnostic_read_dict().keys())))
        return
    return val

def diode_c_read(verbose=True):
    """
    reads the current amplifier and returns the value
    quiet = False to pring
    """ 
    val = diagnostic_read('diode_c',verbose)
    return val

def mesh_d_read(verbose=True):
    """
    reads the current amplifier and returns the value
    quiet = False to pring
    """ 
    val = diagnostic_read('mesh_d',verbose)
    return val

def diode_d_read(verbose=True):
    """
    reads the current amplifier and returns the value
    quiet = False to pring
    """ 
    val = diagnostic_read('diode_d',verbose)
    return val


###############################################################################################
####################################### FLUX CONVERSION #######################################
###############################################################################################




