
import numpy as np
from epics import caget
##############################################################################################################
##############################             Kappa to Euler Conversion        ##############################
##############################################################################################################
def fourc2kappa(theta,chi,phi):
    """
    converts four-circle angles to fourc angles
    **kwargs:
        k_arm: kappa arm angle (default = 50)
        kth_offset: default from EPICS
    """
    _kappa_fourc_calc(theta,chi,phi,'kappa')

def kappa2fourc(kth,kappa,kphi,**kwargs):
    """
    convert kappa anges to fourc angles
    **kwargs:
        k_arm: kappa arm angle (default = 50)
        kth_offset: default from EPICS
    """
    _kappa_fourc_calc(kth,kappa,kphi,'fourc',**kwargs)

def _kappa_fourc_calc(ang1,ang2,ang3,conversion,**kwargs):
    """
    converts kappa angles to four-circle angles
    **kwargs:
        k_arm: kappa arm angle (default = 50)
        kth_offset: from EPICS
    """
    a,b,c = 0,0,0
    if conversion == 'kappa':
        a,b,c = _euler_to_kappa(ang1,ang2,ang3,**kwargs)
        print(("\n"+"th  = "+str(ang1)))
        print(("chi = "+str(ang2)))
        print(("phi = "+str(ang3)))
        print("~~~~~~~~")
        print(("kth  = "+str(a)))
        print(("kap  = "+str(b)))
        print(("kphi = "+str(c)+"\n"))
    elif    conversion=='fourc':
        a,b,c = _kappa_to_euler(ang1,ang2,ang3,**kwargs)
        print(("\n"+"kth  = "+str(ang1)))
        print(("kap  = "+str(ang2)))
        print(("kphi = "+str(ang3)))
        print("~~~~~~~~")
        print(("th  = "+str(a)))
        print(("chi = "+str(b)))
        print(("phi = "+str(c)+"\n"))
    else:
        print('2nd argument invalid; please choose one of the following:')
        print('"Kappa" or  "fourc"')
    return a,b,c



def _euler_to_kappa(e_theta,e_chi,e_phi,**kwargs):
    """
    convert from euler to kappa 
    """
    kwargs.setdefault('k_arm',50)
    kwargs.setdefault('kth_offset',_get_kth_offset)
    k_arm = kwargs['k_arm']
    kth_offset = kwargs['kth_offset']

    pi_deg = np.pi/180.0
   
    if(abs(e_chi)> 2.0*k_arm):
        print("Chi should be no more than twice the Kappa arm offset angle (k_arm = "+str(k_arm)+").")
        kth,kap,kphi=0,0,0
    else:
        k_ang = k_arm*pi_deg
        delta = np.asin( - np.tan(e_chi*pi_deg/2.0) / np.tan(k_ang) )
        
        k_theta = e_theta*pi_deg - delta
        k_kappa = 2.0 * np.asin( np.sin(e_chi*pi_deg/2.0) / np.sin(k_ang))
        k_phi   = e_phi*pi_deg - delta

        kth = np.round_(k_theta/pi_deg)-(57.045-kth_offset)
        kap = np.round_(k_kappa/pi_deg)
        kphi  = np.round_(k_phi/pi_deg)
    return (kth,kap,kphi)

def _kappa_to_euler(k_theta,k_kappa,k_phi,**kwargs):
    """
    convert from kappa to euler
    """
    kwargs.setdefault('k_arm',50)
    kwargs.setdefault('kth_offset',_get_kth_offset)
    k_arm = kwargs['k_arm']
    kth_offset = kwargs['kth_offset']
       
    pi_deg = np.pi/180.0
    k_ang = k_arm*pi_deg
    delta = np.atan( np.tan(k_kappa*pi_deg/2.0) * np.cos(k_ang) )
    e_theta = k_theta*pi_deg - delta
    e_chi   = 2.0 * np.asin( np.sin(k_kappa*pi_deg/2.0) * np.sin(k_ang) )
    e_phi   = k_phi*pi_deg - delta

    theta = np.round_(e_theta/pi_deg,3)+(57.045-kth_offset)
    chi   = np.round_(e_chi/pi_deg,3)       # convert from rad to deg
    phi   = np.round_(e_phi/pi_deg,3)
    return (theta,chi,phi)


def _get_kth_offset(verbose=True):
    try:
        kth_offset=caget('29idKappa:userCalcOut1.G')
    except:
        kth_offset = 0 
    if verbose:
        print(("Calculating Euler angles using kth0 = "+str(kth_offset)))
    return kth_offset