from os.path import join, isfile, exists, dirname
from time import sleep
import matplotlib.pyplot as plt

from epics import caget, caput

import  iexcode.instruments.cfg as iex

from iexcode.instruments.xrays import *
from iexcode.instruments.VLS_PGM import *
from iexcode.instruments.IEX_VPU import *

from iexcode.instruments.current_amplifiers import *
from iexcode.instruments.diagnostics import diagnostics_all_out, diagnostics_all_in,diode_c,diode_d
from iexcode.instruments.m3r import *
from iexcode.instruments.slits import _slits_wide_open_dictionary
from iexcode.instruments.slits import *
from iexcode.instruments.utilities import *
from iexcode.instruments.electron_analyzer import scanEA
from iexcode.macros.cam_macros import *

from .quick_plot import fit_mda, mda_1D

##############################################################################################################
################################            checking alignment             ##############################
##############################################################################################################

def check_flux_scan(hv=500,stay=False,**kwargs):
    """
    puts the diode in and scans the mono
    stay: 
        True => leaves the diode in
        False => takes the diode out
    """
    kwargs.setdefault('sound',True)
    diagnostics_all_out(diode_stay_in=True)
    slit(50)
    energy(hv)
    hv_start = hv*.95
    hv_stop = hv*1.05
    hv_step = round(1.0*hv/500,0)
    scanmono(hv_start,hv_stop,hv_step,comment='Mono Scan @ '+str(hv)+' eV')
    if kwargs['sound']:
        playsound()

def check_flux_read(hv=500,ID_mode='RCP',stay=False):
    """
    puts the diode in in the branch defined my position of m3r
    measures flux at energy ID_mode specified
    stay = True => leaves the diode in

    Previously: CheckFlux
    """
    current_slit = exit_slit_get(verbose=True)
    ID_switch_mode(ID_mode)
    energy(hv)
    
    ring_current=round(caget("S:SRcurrentAI.VAL"),2)

    diagnostics_all_out(diode_stay_in=True)
    exit_slit_get(50)
    sleep(10)

    if m3r_branch() == "c":
        diode=diode_c_read()
    elif m3r_branch() == "d":
        diode=diode_d_read()

    flux=current2flux(diode)
    print("\n----- Current on diode   : %.3e" % diode, "A")
    print("----- Corresponding flux: %.3e" % flux, "ph/s \n")
    print("----- Storage ring current: %.2f" % ring_current, "mA")
    print("\nFlux at hv=500 as off Feb 2019 for RCP and HEG:  ~3.3e-06 A = ~1.5e+11 ph/s")
    
    if not stay:
        diagnostics_all_out()
    exit_slit_set(current_slit)


def scan_wire(direction,all_diag=True,**kwargs):
    """
    Scans the wires located just downstream of M0/M1, 
         which = 'H' for the horizontal, typically CA2
         which = 'V' for the vertical, typically CA3
    all_diag =True -> AllDiagIn(), otherwise you have to put any diagnostics in by hand
    Logging is automatic: use **kwargs or the optional logging arguments see scanlog() for details    
    **kwargs:
        execute
    
    Previously: WireScan
    """
    kwargs.setdefault('execute',True)
    if all_diag:
        diagnostics_all_in()
    
    
    if direction == 'H':
        name = "H-wire"

    elif direction == 'V':
        name = "V-wire"

    diagnostics_presets_go(name,'Start')
    pv = diagnostics_presets()[name][1]
    print("\n================== "+name+" scan (29idb:ca2):")
    iex.BL.mda.fillin(pv+".VAL",pv+".RBV",1,-17,-30,-0.25,**kwargs)

    if kwargs['execute']:
        iex.BL.mda.go(**kwargs)

    if all_diag:
        diagnostics_all_out()


##############################################################################################################
###########################                    Beam Profile               ######################
##############################################################################################################
def scan_narrow_slit(slit='2V',slit_parameters=[0.25,-2,2,0.5],**kwargs):
    """
        slit = '1V','1H','2V','2H'
        slit_parameters = [SlitSize,start,stop,step]
  
    Typical slit sizes/start/stop/step are (for full range): 
        1H/1V : [0.50, -4.5, 4.5, 0.2]
        2H    : [0.25, -3.0, 3.0, 0.2]
        2V-MEG: [0.25, -4.0, 4.0, 0.2]    
        2V-HEG: [0.50, -8.0, 8.0, 0.2] 

    **kwargs:
        scan_dim = 1
        execute = True

    Previously: Scan_NarrowSlit
    """
    kwargs.setdefault('scan_dim',1)
    kwargs.setdefault('execute',True)

    size,start,stop,step = slit_parameters
      
    direction = slit[1]
    if direction == "V":
        size = (inf, size) 
        center = (0, 0)
    elif direction == "H":
        size = (size, inf) 
        center = (0, 0)

    if slit[0] == '1':
        slit_name = "slit1A"
        slit1A_set(size,center)
        slit2B_set((inf,inf),(0,0))

    elif slit[0] == '2':
        slit_name = "slit2B"
        slit1A_set((3,3),(0,0))
        slit2B_set(size,center)
  
    slits_scan_center(slit_name,direction,start,stop,step,**kwargs)
    if kwargs['execute']:
        iex.BL.mda.go(kwargs['scan_dim'])


def scan_mono_vs_slit(slit='2V',slit_parameters=[0.25,-2,2,0.5],energy_parameters=[470,530,2],**kwargs):
    """
    This can be used to find the center of the resonant beam i.e. the slit value for the most blue shifted curve: 
        slit='1V','1H','2V','2H'
        slit_parameters = [SlitSize,start,stop,step]
        energy_parameters = [eVstart,eVstop,eVstep]= [470,530,2]
        
    Typical slit sizes/start/stop/step are (for full range): 
        1H/1V : [0.50, -4.5, 4.5, 0.2]
        2H    : [0.25, -3.0, 3.0, 0.2]
        2V-MEG: [0.25, -4.0, 4.0, 0.2]    
        2V-HEG: [0.50, -8.0, 8.0, 0.2] 

    Previously: Scan_MonoVsSlit
    """
    hv_start,hv_stop,hv_step = energy_parameters  
    # Filling Scans:
    mono_scan(hv_start,hv_stop,hv_step)
    scan_narrow_slit(slit='2V',slit_parameters=[0.25,-2,2,0.5],**kwargs)

    # Resetting everybody to normal:
    mono_energy_set((hv_start+hv_stop)/2.0)
    apertures_set()


##############################################################################################################
################################            mono alignment             ##############################
##############################################################################################################
def mono_MIR_GRT_find_offset(grt,slit_list,**kwargs):    
    """ 
    Find MIR-GRT offset by scanning 0 order through exit slit
        **kwargs:
            detNum = 15

    Previously: MIR_GRT_Offset
    """
    kwargs.setdefault("detNum",15)

    grating(grt)
    exit_slit_set(50)
    diode_c('In')
    slit2B_set(2,0.5,0,0)
    
    #set zero order, scan grating pitch
    for ang in range_up(1.5,5,0.5):
        print("\r")
        mono_zero_order(ang)
        print("Mirror pitch: "+str(ang)+"\r")
        for s in slit_list:
            exit_slit_set(s)
            print("\r")
            mono_motor_scan("GRT:P",ang-0.0005*s, ang+0.0005*s ,0.00002*s)

            sleep(1)
            print("\r")
            grating_ang = fit_mda(iex.BL.mda.fileNum(),kwargs['detNum'],.1,'gauss')
            print("Peak: ",grating_ang)
            print("\r")
            print("-------------------------")

##################### mono slit scans => Energy calibration #####################
def exit_slit_vs_hv(hv,c=3):
    '''
    Adjust slit size to keep reasonable count vs hv
          c=3 for nominal slit2B size
          c=10 for apertured slit2B (e.g 0.25 or 0.5)

    Previously: Mono_Set_Slitvshv
    '''
    slit(hv/100.0*c)


def scan_energy_along_grating(hv,peakBE=84,pts=11,r=0.75,i=1,**kwargs):
    """
    Takes a Slit-2V map for a range of photon energies 
        peakBE=84; fixed mode
        pts = number of pts in slit positions, default: 11
        r = % of range in slit size,  default: 0.75
        i = scanEA time in minute (default:1 for MEG, 2 for HEG?)
        Takes 2.5h with default parameters (i=1)

    Previously: Mono_Scan_Slit2BV  
    """
    grt_density = mono_grating_density_get()
    c=grt_density/1200                # c = 1 for MEG, 2 for HEG

    # Getting EA parameters vs hv:
    PE=200
    
    #Getting Slit-2B parameters
    Vsize=0.25*c
    n=(4.5-hv/500)
    Vstart=-((floor((2*n)/Vsize/2)+1)*Vsize*c)*r
    Vstep=-round(Vstart*2/pts,2)

    #Setting energy:
    energy(hv)
    exit_slit_vs_hv(hv,c=10)

    #Aperturing slit 2V (note that we have the nominal size horizontally)
    (size_rbv,size_val),(center_rbv,center_val) = slits_pvs('slit2B')
    caput(size_val,Vsize)
    
    print("\n--------------- Starting core level vs Slit-2B(V) map ---------------")

    #Take XPS for each Slit-2V position
    for V_center in range_up(Vstart,-Vstart,Vstep):
        caput(center_val,V_center,wait=True,timeout=18000)
        kwargs.update({'comment': "2-V center ="+str(V_center)[0:6]})
        EAlist=["BE",peakBE,PE,17*60*i,1]
        scanEA(EAlist,**kwargs)
    
    #reset the slits
    apertures_set()



def scan_mono_energy_drift(start,stop,step,peakBE=84,EF=None,**kwargs):
    '''
    Measure core level (by default Au 4f) in fixed mode vs energy while maintaining resonable count rate
    If EF='y' (by default None) takes the Fermi edge as well.
    kwargs:
        c=3 is the default slit2B size
    
    Used for grating calibration (changing grating_offset and b2 effects this - use grating_offset since b2 was measured)
    
    Previously: Mono_Scan_hv_drift
    '''
    kwargs.setdefault('c','3')

    for hv in range_up(start,stop,step):
        energy(hv)
        c=int(kwargs['c'])
        exit_slit_vs_hv(hv,c)
        kwargs.update({'comment': "Core level @ hv = "+str(hv)})      
        scanEA(["BE",peakBE,200,17*60*1,1],**kwargs) 
        if EF is not None:
            kwargs.update({'comment': "EF @ hv = "+str(hv)})                
            scanEA(["BE",0,200,17*60*10,1],**kwargs) 


    
##############################################################################################################
###########################                    Pinhole Scripts               ######################
##############################################################################################################
def scan_pinhole_full():
    """
    Does a pinhole with .1 mm step with AllDiagIn and then with only branch diode in
    
    Previously: FullPinhole
    """
    diagnostics_all_in()
    scan_pinhole([-3, 3, .5, .1], [-3, 3, .1, .5])
    diagnostics_all_out(diode_stay_in=True)
    scan_pinhole([-3, 3, .5, .1], [-3, 3, .1, .5])


def scan_pinhole(H_list,V_list,**kwargs):
    """
    Sets up the pinhole scan for slit1A
    Make sure that you have the appropriate Diagnostics in can use AllDiagIn()

    H_list = [Hstart, Hstop, Hstep, Hsize] => [-3, 3, .1, .5]
    V_list = [Vstart, Vstop, Vstep, Vsize] => [-3, 3, .1, .5]
    
    *kwargs:
        H_scan_dim => 1
        V_scan_dim => 2
        execute => True

    Previously: Scan_Pinhole_Go, Scan_Pinhole
    """
    kwargs.setdefault('H_scan_dim',1)
    kwargs.setdefault('V_scan_dim',2)
    kwargs.setdefault('execute',True)

    #Synching slits
    slits_synch("slit1A")

    #Getting initial slit size
    (Hsize0,Vsize0),(Hcenter0,Vcenter0) = slit1A_get(verbose=False)

    #Filling in the ScanRecord
    slits_scan_center("slit1A","H",H_list[0],H_list[1],H_list[2],scan_dim=kwargs['H_scan_dim'],execute=False)
    slits_scan_center("slit1A","V",V_list[0],V_list[1],V_list[2],scan_dim=kwargs['V_scan_dim'],execute=False)

    #Set the slit size
    slit1A_set((H_list[2],V_list[2]),(0,0),verbose=True)

    if kwargs['execute']:
        print("\rPinhole scan: ("+str(kwargs['H_scan_dim'])+","+str(kwargs['V_scan_dim'])+") "
        +str(H_list[0])+"/"+str(H_list[1])+"/"+str(H_list[2])+"/"+str(H_list[3])+"/"
        +str(V_list[0])+"/"+str(V_list[1])+"/"+str(V_list[2])+"/"+str(V_list[3]))

        iex.BL.mda.go(max(kwargs['H_scan_dim'],kwargs['V_scan_dim']))
        print("\rNow setting the slits back to:")
        apertures_set()

        print("\r")
        print_warning_message("Don't forget to pull all of the diagnostics out.")
   


##############################################################################################################
##################             Slit Calibration & Beam Steering scripts                 ##################
##############################################################################################################
FMB_mirror_move

def notes_20191212():
    print(""" looking at diode 3 in scattering chamber, open 3D all the way (2000)
        WARNING: 10% energy detune is enough to see the 2 lobes with 2B but not for 1A => 20% 
        Red shifted  (485eV) = edges give center of the grating
        Blue shifted (510eV) = gaussian center gives center of the beam
        => we want the slit centered on the grating center and the beam centered on the slits/grating
        => 5 urad down = 1 mm negative motion on 2B-V with HEG""")


def procedure_align_m1():
    print("""\n-Start with a very red shifted line cut to localize the edges of the grating:
                Set_IDraw(510)
                mono(485)
                scan_narrow_slit('2V',[0.25,-8,8,0.5])
                scan_narrow_slit('2H',[0.25,-4,4,0.5])
    => Verticaly: The edges/center of the grating define the center of 2V 
    => Horizontaly: The edges are defined by the aperture upstream. 
        As long as the beam makes it through the pipe to the next mirror, we are good. 
        Each M1 pitch will have its own slit-2H zero
        It can be a good idea to check that GRT or M2 translation does not change anything.
\n-Once the slits are centered on the grating, center resonant beam on the grating (zero slit):
                FMB_mirror_get(1)   # to get current values
                for roll_value in RangeUp(3,4,0.2):
                    print('\\nRZ = '+str(roll_value))
                    FMB_mirror_move(1,'RZ',roll_value)
                    sleep(5)
                    scan_mono_vs_slit('2V',[0.25,-0.5,0.5,0.25],[490,530,2])
                for pitch_value in RangeUp(8.5,8.7,0.05):
                    print('\\nRY = '+str(pitch_value))
                    FMB_mirror_move(1,'RY',pitch_value)
                    sleep(5)
                    scan_mono_vs_slit('2H',[0.25,-0.5,0.5,0.25],[490,530,2])
    => The resonant beam is the slit value for the most blue shifted curve.
    
    WARNING: It is more accurate (and easier) to compare the leading edge of the ID peak 
    than trying to look at the perpendicular direction (slit line cut) because of the 
    asymetry in the beam intensity (uneven carbon strip along the grating?)
""")


    # 2017_3 - HEG after opitmization:  mba_b_0241-0244
    # 2018_1 - HEG before opitmization: mba_b_0005-0008

def iex_steering(h_position,v_position):
    """
    prints the required steering to compensate for the current
    """
    print('\nBeam position:')
    print('   - horizontal: '+str(h_position))
    print('   - vertical:   '+str(v_position))
    h_steering,v_steering=beamsteering_1A(h_position,v_position)
    print('\nBeam steering:')
    print('   - horizontal: '+h_steering)
    print('   - vertical:   '+v_steering)
    print('\n')
    return h_steering,v_steering

def beamsteering_1A(h_position,v_position,FE_center_h=0,FE_center_v=0):
    h_correction=FE_center_h-h_position
    v_correction=FE_center_v-v_position
    hrad=round(abs(h_correction)*10/0.25,0)
    vrad=round(abs(v_correction)*10/0.25,0)
    if h_correction<0: h_direction=' urad outboard'
    else: h_direction=' urad inboard'
    if v_correction<0: v_direction=' urad up'
    else: v_direction=' urad down'
    h_steering=str(hrad)+h_direction
    v_steering=str(vrad)+v_direction
    return h_steering,v_steering

def M0M1_steering(H2,V2):
    """
        M1 steering: (monoVslit-2V and 2H)
        Roll:  Rz more positive => 2V more positive (0.25 => .25 mm)
        Pitch: Ry more positive => 2H more positive (0.03 => .5 mm
    """
    print('\nBeam position:')
    print(f"   - vertical:   V2 = {round(V2,3)}")
    print(f"   - horizontal: H2 = {round(H2,3)}")
    v_steering=beamsteering_2B(V2)
    m1_roll=beamsteering_2B(V2)[1]
    print('\nCurrent position:')
    current_roll=FMB_mirror_get(1)[5]
    new_roll=current_roll+m1_roll
    print('\nREMEMBER: 2V slit should always be centered at zero by steering M1 roll!')
    print("   - vertical: "+str(current_roll)+str(v_steering[1])+" => Move_M1('RZ',"+str(new_roll)+")")
    print('   - horizontal: reset slit 2BH center to ',str(H2))  

    print("\n\nM1 roll vs energy calibration/flux:")
    print("   - 1.0 mm @ slit2V ~ 1.0 mrad on M1 roll")
    print("   - 1.0 mm @ slit2V = 0.4 eV shift in energy")
    print("   - 0.8 mm @ slit2V = 25% loss in intensity on Au 4f\n")

         
def beamsteering_2B(v_position):
    vroll=round(v_position*0.05/0.04,3)
    if v_position<0: v_direction=' +'
    else: v_direction=' -'
    v_steering=v_direction+str(abs(vroll))+' urad'
    return v_steering,float(v_steering.split(" ")[1])



   

def ID_steering_scans(**kwargs):    
    """
    slit1 - Scanning front-end slit center
    cutoffs are defined by the fixed aperature midpoint is slit center
    beam should be center at slit center, steer ID until the coincide

    **kwargs:
        verbose: default => True
        sound:  playsound, default => True 
        extended_range True/False
            False: default range = -3,3,0.1
            True: -5,5,0.1

    Note: Does not set the ID, mvID = 2.0 gives a nice gaussian
    """
    kwargs.setdefault('sound',True)

    if 'extended_range' in kwargs:
        start,stop = (-5,5)
    else:
        start,stop = (-3,3)

    size,center = slit1A_get(verbose=False)

    slits_scan_center('slit1A','H',start,stop,0.1,size=(0.25,inf),comment='Slit center - 1H')
    slits_scan_center('slit1A','V',start,stop,0.1,size=(inf,0.25),comment='Slit center - 1V')
    
    slit1A_set(size,center,verbose=False)


    if kwargs['sound']:
        playsound()

    if kwargs['verbose']:
        print('\nsteering out => move beam more positive (10 urad ~ 0.25 mm)')
        print('steering up  => move beam more positive (10 urad ~ 0.25 mm)\n')

def ID_steering_plots(scanNum,det_num,xrange=[-1,1]):
    """
    plots data from ID_steering_scans and returns the positions
    """
    
    position=fit_mda(scanNum,det_num,1,'gauss',xrange=xrange)
    plt.show()

    return position

def wire_scans(**kwargs):
    """
    Checks for large misaligment or clipping by M0/M1

    Note: Does not set the ID, mvID = 2.0 gives a nice gaussian
    """
    kwargs.setdefault('sound',True)

    slits_set_size('slit1A',(4.5,4.5))
    scan_wire('H',comment='Wire scan - H',all_diag=False)     # by default puts all meshes in + DiodeC
    scan_wire('V',comment='Wire scan - V',all_diag=False)

    if kwargs['sound']:
        playsound()

def monoVslit_quick_scan(slit_name, **kwargs):
    """
    sets the slit to be narrow and scans along and across the grating
    at 500 eV to see if M0/M1 are steering the beam  to the center of the grating

    """
    kwargs.setdefault('sound',True)

    energy(500)
    mvmono(505)

    H,V = _slits_wide_open_dictionary()[slit_name]

    c = mono_grating_density_get()/1200.0

    slits_scan_center(slit_name,'V',-V,V,0.25,size=(inf,0.25*c))
    sleep(1)

    slits_scan_center(slit_name,'H',-H,H,0.25,size=(0.25,inf))
    sleep(1)
        
    if kwargs['sound']:
        playsound()


def monoVslit_quick_plot(scanNum,detDiode,c,xrange=None):
    position = fit_mda(scanNum,detDiode,c,'gauss',xrange)
    plt.show()
    return position


#####################################################################################
### Slit-1A Procedures
#####################################################################################

def check_slit1A(step=0.1):
    """
    Checks that Slit1A is centered on the fixed aperature/theoretical optic axis
        top='m9' -> CA5
        inboard='m10' -> CA4
        
    Scans the top blade - CA4 sees an step function
           top_center = midpoint of motor position for zero current and when current saturates
    Scans the inner blade - CA4 sees box fuction convoluted with gaussian beam position 
           inner_center - midpoint bewteen box edges
           
    SetSlit1A(0,0,inner_center,top_center)
    for m in range(9,13):
        Reset_Motor_User(m,'b',0)
    SyncAllSlits()  
    
    If data makes no sense then Reset_Slit1A_Procedure()
 
    """
   #scanning top-blade
    slit1A_set(8,8,0,0)         # aperture wide open
    caput('29idb:m10.VAL',0)   # Inboard blade centered-ish in the beam
    m=9
    VAL='29idb:m'+str(m)+'.VAL'
    RBV='29idb:m'+str(m)+'.RBV'
    iex.BL.mda.fillin(VAL,RBV,-4,4.0,step)
    FileNum1  = iex.BL.mda.lastFileNum()

    #scanning inboard-blade
    slit1A_set(8,8,0,0)         # aperture wide open
    caput('29idb:m9.VAL',0)    # top blade centered-ish  in the beam
    m=10
    VAL='29idb:m'+str(m)+'.VAL'
    RBV='29idb:m'+str(m)+'.RBV' 
    iex.BL.mda.fillin(VAL,RBV,-4,4.0,step)
    FileNum2  = iex.BL.mda.lastFileNum()

    return FileNum1, FileNum2


def procedure_reset_slit1A():
    """
    Prints the procedure for Resetting Slit1A
    """
    print("\n#------- Checking and Resetting the Dial -------")
    print("    GoToSlit1AScribe()")
    print("# if not on the scribe marks tweek the position until the scribe marks are aligned")
    print("    for m in [9,10,11,12]: Reset_Motor_Dial(m,\'b\',0);SyncAllSlits()")
    print("\n#------- Resetting the User -------")
    print("# SetSlit1A by hand such that you have some beam coming through")
    print("    Scan_SlitSize(\"1H\",start,stop,step) #looking at CA4")
    print("    Scan_SlitSize(\"1V\",start,stop,step) #looking at CA4")
    print("# Then SetSlit1A(1Hsize,1Vsize,0,0); where 1Hsize and 1Vsize are the size where CA4 = 0")
    print("    for m in [9,10,11,12]: Reset_Motor_User(m,'b',0);SyncAllSlits()")
    
  
def check_ID_steering(hv=2000):
    """
    Scans Slit1A center (set to a (0.25,0.25) pinhole) while looking at the back blade:
        - slit center vs fixed aperture: given by the position of the edges
        - beam center vs fixed aperture: given by the position of the bump in the middle
    """
    slit1A_set(0.25,0.25,0,0)
    ID_set(hv)
    scan_slit_center('1H',-3,3,0.1)
    scan_slit_center('1V',-3,3,0.1)
        
def scan_slit_center(slit,start,stop,step,size=0.25):
    """
    slit = '1H','1V','2H' or '2V'

    Previously: Scan_SlitCenter
    """
    if slit[0] == '1':
        slit_name = "slit1A"
    elif slit[0] == '2':
        slit_name = "slit2B"
    direction = slit[1]
    slits_scan_center(slit_name,direction,start,stop,step,size=size)


def scan_slit_size(slit,start,stop,step,center=(0,0)):
    """
    slit = '1H','1V','2H' or '2V'

    Previously: Scan_SlitSize
    """
    if slit[0] == '1':
        slit_name = "slit1A"
    elif slit[0] == '2':
        slit_name = "slit2B"
    direction = slit[1]
    slits_scan_center(slit_name,direction,start,stop,step,center=center)



#####################################################################################
# Checking the beam steering uses the gas-cell to see if the beam is centered on the grating
# and that the slits are centered
#####################################################################################

def beam_profile(grt,slit_list):
    """
        Makes a nice 2D image of the energy distribution of the beam across the grating at ID=500
        SlitList=["2H","2V","1H","1V"]

        Note: Does NOT put the diagnostics into the beam 

        Previously BeamProfile
    """
    grating(grt)
    c = mono_grating_density_get()/1200
    
    mvID(500)
    apertures_set()

    eVstart = 460
    eVstop = 540
    eVstep = 4

    for slit in slit_list:
        if slit=="1H":
            #Hsize,Vsize,Hstart,Hstop,Hstep = 0.50,2,-2,2,0.25*c_slit      # => 35 min
            scan_mono_vs_slit('1H',[0.5,-2,2,0.25*c],[eVstart,eVstop,eVstep],comment='Mono/Slit - 1H')
        elif slit == "1V":
            #Hsize,Vsize,Vstart,Vstop,Vstep = 2,0.50,-2,2,0.25*c_slit
            scan_mono_vs_slit('1V',[0.5,-2,2,0.25*c],[eVstart,eVstop,eVstep],comment='Mono/Slit - 1V')
        elif slit =="2H":
            #Hsize,Vsize,Hstart,Hstop,Hstep = 0.50,8,-3,3,0.5*c_slit
            scan_mono_vs_slit('2H',[0.5,-3,3,0.5*c],[eVstart,eVstop,eVstep],comment='Mono/Slit - 2H')
        elif slit =="2V":
            #Hsize,Vsize,Vstart,Vstop,Vstep = 6,0.25*c,-4*c,4*c,0.5*c_slit
            scan_mono_vs_slit('2V',[0.5,-4,4,0.5*c],[eVstart,eVstop,eVstep],comment='Mono/Slit - 2V')


def CheckAllSlits_long(hv_list=[500],slit_list=["1H","1V","2H","2V"],**kwargs):
    """
    For each photon energy in hvList, takes 3 slit curves @ mono below / resonance / above
    For given SlitList
    For both gratings
    """
    for grt in ["HEG","MEG"]:    # One full loop for a given grating and 3 energies takes 5h for average 20
        grating(grt)
        for hv in hv_list:
            apertures_set(hv)
            step=hv/100.0
            start=hv-step
            stop=hv+step
            for slit in slit_list:
                for hv in range(start,stop,step):
                    print("\r")
                    mono_energy_set(hv)
                    check_ID_beam_steering(slit)


def CheckSlitCalibration(slit_list,BL=500,hvList=[485,510],scanIOC=None,**kwargs):
    """
    Slit scan for red shifted and blue shifted mono values
    used to determine in the beam is centered on the grating (gaussian)
    and if the slits are centered on the grating (humped)
                 hv=500; hvList=[485,510]
                 hv=1500;hvList=[hv*0.97,hv*1.01]
    Note: sets the exit slit to 50
    """
    if scanIOC is None:
        scanIOC=BL_ioc()
    exit_slit_set(50)
    apertures_set(BL)
    for hv in hvList:
        mono_energy_set(hv)
        for slit in slit_list:
            check_ID_beam_steering(slit)
        print("\n")

def check_ID_beam_steering(slit):   #here
    """
    checks the ID beam steering by scanning the slit with the ID detuned
    if there is carbon on the mirror you might need to do the full ID_beam_profile to
    find the red shifted beam

    slit = '1V','1H','2V','2H'

    Previously: CheckBeamPosition
    """
             
    grt_density = mono_grating_density_get()    # can be plotted directly on dview using the buffer. That should
    c=grt_density/1200                # be fairly quick and not require any data loading/analysis
    
    slit1A_set((3.5,3.5),(0,0),verbose=False)
    slit2B_set((6.0,8.0),(0,0),verbose=False)        # Open up all the slits to make sure we scan the full beam
    
    if slit == "1H":
        size,start,stop,step = 0.50,-4.5,4.5,0.2
    elif slit == "1V":
        size,start,stop,step = 0.50,-4.5,4.5,0.2
    elif slit == "2H":
        size,start,stop,step = 0.25,-8.0,8.0,0.2
    elif slit == "2V":
        size,start,stop,step = 0.25*c,-4*c,4*c,0.2*c

    scan_slit_center(slit,start,stop,step,size=size)    
    apertures_set()


##############################################################################################################
##################                 Commissioning paper curves                     ##################
##############################################################################################################
def FermiEdges(Energy,Slit):
    EF_Table={}
    EF_Table[500]  = {10:20,  20:20,  50:20,  100:20,  200:20}
    EF_Table[1000] = {10:50,  20:50,  50:50,  100:50,  200:50}
    EF_Table[1500] = {10:100, 20:100, 50:100, 100:100, 200:100}
    PE = EF_Table[Energy][Slit]
    return PE




def QP_curves(ID_energy,ID_mode_list,hv_start,hv_stop,hv_step):
    """
    Sets the ID to setpoint and scans the mono for a list of polarizations

    Does not set the QP_ratio first use, ID_QP_mode_set(QP_ratio)
    """
    exit_slit_set(200)

    for ID_mode in ID_mode_list:
        polarization(ID_mode)
        ID_set_eV(ID_energy)
        apertures_set()
        mono_scan(hv_start,hv_stop,hv_step)





##############################################################################################################
##############################             Beam Motion            ##############################
##############################################################################################################
def scan_cam_motor(motor_name,start,stop,step,cam_num,ExposureTime):
    """   
    motor_name = 'x','y','z'...
    cam_num = 1,2,3... uses cam_pv to get the pv

    Previously: Pixel_Calibration
    """    

    cam_scan_setup(cam_num,ADtype='TIFF',)

    iex.BL.motors.scan(motor_name,start,stop,step)
    cam_live()



def scan_cam_hv(start,stop,step,cam_num,**kwargs):
    """   
    takes and image as a function of photon energy
    **kwargs:
        ExposureTime
    Previously: BeamMotion_Vs_BL_Go
    """
    cam_scan_setup(cam_num,ADtype='TIFF')
    scanmono(start,stop,step,**kwargs)

    cam_live()



def scan_cam_mono_zero_order(start,stop,step,cam_num,iterations=1,**kwargs):
    """
    Previously: BeamMotion_Vs_MonoZero_Go
    """
    i = 0
    while i < iterations:
        for angle in range(start,stop+step,step):
            mono_zero_order(angle)
            sleep(5)
            cam_snap(cam_num,ADtype='TIFF',**kwargs)
        i+=1
    
    cam_live()

def scan_cam_mono_pitch(motor,angle,delta,cam_num,iterations=1,**kwargs):
    """
    takes image at zero order (alpha = beta = angle)
    takes image at zeror order with delta 

    motor = 'grating' or 'mirror
        if 'grating': 
            alpha = angle
            beta = angle + delta 
        if 'mirror': 
            alpha = angle + delta 
            beta = angle

    Previously:BeamMotion_Vs_MirPitch_Go
    """
    if motor == 'grating': 
        alpha = angle
        beta = angle + delta 
    if motor == 'mirror': 
        alpha = angle + delta 
        beta = angle
    i=0
    while i<iterations:
        mono_angles_set(angle,angle)
        sleep(5)
        cam_snap(cam_num,ADtype='TIFF',**kwargs)
        
        mono_angles_set(alpha,beta)
        sleep(5)
        cam_snap(cam_num,ADtype='TIFF',**kwargs)
        i+=1

    cam_live()



